/*
 * LCD3WI.cpp
 *
 *  Created on: 10 сент. 2016 г.
 *      Author: alexandr
 */

#include "LCD3WI.h"

#include <avr/pgmspace.h>
#include <avr/eeprom.h>



static uint8_t _registerData = 0;

static uint8_t _displayfunction = LCD_4BITMODE | LCD_5x8DOTS | LCD_2LINE | LCD_PAGE2;
static uint8_t _displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;
static uint8_t _displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;

static uint8_t _row_offsets[2] = { 0x00, 0x40 };



void LCD_write4bits(uint8_t value);
void LCD_setupDisplayPins();
void LCD_send(uint8_t value, uint8_t mode);
void LCD_pulseEnable();


void LCD_begin()
{

	delayMicroseconds(50000);

	LCD_setupDisplayPins();

	LCD_write4bits(0x03);
	delayMicroseconds(4500);

	LCD_write4bits(0x03);
	delayMicroseconds(4500);

	LCD_write4bits(0x03);
	delayMicroseconds(150);

	LCD_write4bits(0x02);

	LCD_send(LCD_FUNCTIONSET | _displayfunction, 0);
	LCD_send(LCD_DISPLAYCONTROL | _displaycontrol, 0);
	LCD_clear();
	LCD_send(LCD_ENTRYMODESET | _displaymode, 0);

}

void LCD_printChar(char ch)
{
	LCD_send(ch, 1);
}

void LCD_printNumber(uint8_t n)
{
	LCD_printChar((n / 10) + '0');
	LCD_printChar((n % 10) + '0');
}

void LCD_printTime(uint8_t m, uint8_t s)
{
	LCD_printNumber(m);
	LCD_printChar(':');
	LCD_printNumber(s);
}

void LCD_printUTF8String(const char *utf8String)
{
	for (size_t i = 0; utf8String[i] != 0; i++) {
		if ((utf8String[i] == 0xD0) || (utf8String[i] == 0xD1)) {
			uint16_t ch = utf8String[i++] * 0x100;
			ch += utf8String[i];
			LCD_printUTF8Char(ch);
		} else {
			LCD_printChar(utf8String[i]);
		}
	}
}

void LCD_printUTF8Char(uint16_t ch)
{
	if (ch == 0xD081) { // letter 'Ё'
		LCD_printChar((char) 0xA8);
	} else if (ch == 0xD191) { // letter 'ё'
		LCD_printChar((char) 0xB8);
	} else if (ch > 0xD100) {
		ch &= 0xFF;
		LCD_printChar(ch + 0x70);
	} else if (ch > 0xD000) {
		ch &= 0xFF;
		LCD_printChar(ch + 0x30);
	} else {
		LCD_printChar((char) ch);
	}
}

void LCD_setupDisplayPins()
{
	for (uint8_t i = 0b10000000; i > 0; i >>= 1) {
		if (i & _registerData)
			PORTB |= 0x01;
		else
			PORTB &= 0xFE;
		PORTB |= 0x04;
		delayMicroseconds(10);
		PORTB &= 0xFA;
	}
	LATCH_SHREG
}

uint8_t swapbits(uint8_t value)
{
	uint8_t p1 = value & 0x03;
	p1 |= p1 * 4;	// smaller code than p1 |= p1 << 2;
	p1 <<= 1;
	p1 &= 0x0C;
	uint8_t p2 = value & 0x0C;
	p2 |= p2 / 4;	// smaller code than p2 |= p2 >> 2;
	p2 >>= 1;
	p2 &= 0x03;
	return (p1 | p2) & 0xF;
}

void LCD_write4bits(uint8_t value)
{
	_registerData &= 0xE0;
	_registerData |= swapbits(value) << 1;
	LCD_setupDisplayPins();
	LCD_pulseEnable();
}

void LCD_noBlink()
{
	_displaycontrol &= ~LCD_BLINKON;
	LCD_send(LCD_DISPLAYCONTROL | _displaycontrol, 0);
}

void LCD_blink()
{
	_displaycontrol |= LCD_BLINKON;
	LCD_send(LCD_DISPLAYCONTROL | _displaycontrol, 0);
}

void LCD_clear()
{
	LCD_send(LCD_CLEARDISPLAY, 0);
	delayMicroseconds(2000);
}

void LCD_setCursor(uint8_t col, uint8_t row)
{
	LCD_send(LCD_SETDDRAMADDR | (col + _row_offsets[row]), 0);
}

void LCD_send(uint8_t value, uint8_t mode)
{
	_registerData &= 0xBF;
	_registerData |= 0x40 * mode;

	LCD_setupDisplayPins();

	LCD_write4bits(value >> 4);
	LCD_write4bits(value);
}

void LCD_pulseEnable()
{
	_registerData &= 0xDF;
	LCD_setupDisplayPins();
	delayMicroseconds(1);

	_registerData |= 0x20;
	LCD_setupDisplayPins();
	delayMicroseconds(1);

	_registerData &= 0xDF;
	LCD_setupDisplayPins();
	delayMicroseconds(100);
}

