/*
 * LCD3WI.h
 *
 *  Created on: 10 сент. 2016 г.
 *      Author: alexandr
 */

#ifndef LCD3WI_H_
#define LCD3WI_H_


#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include "Utils.h"
#include "Config.h"


// settings
#define LCD_NUMLINES	2
#define LCD_NUMCOLUMNS	16

#define SR_DB4		4
#define SR_DB5		3
#define SR_DB6		2
#define SR_DB7		1

#define SR_ENABLE	5
#define SR_RESET	6


// commands
#define LCD_CLEARDISPLAY 0x01
#define LCD_RETURNHOME 0x02
#define LCD_ENTRYMODESET 0x04
#define LCD_DISPLAYCONTROL 0x08
#define LCD_CURSORSHIFT 0x10
#define LCD_FUNCTIONSET 0x20
#define LCD_SETCGRAMADDR 0x40
#define LCD_SETDDRAMADDR 0x80

// flags for display entry mode
#define LCD_ENTRYRIGHT 0x00
#define LCD_ENTRYLEFT 0x02
#define LCD_ENTRYSHIFTINCREMENT 0x01
#define LCD_ENTRYSHIFTDECREMENT 0x00

// flags for display on/off control
#define LCD_DISPLAYON 0x04
#define LCD_DISPLAYOFF 0x00
#define LCD_CURSORON 0x02
#define LCD_CURSOROFF 0x00
#define LCD_BLINKON 0x01
#define LCD_BLINKOFF 0x00

// flags for display/cursor shift
#define LCD_DISPLAYMOVE 0x08
#define LCD_CURSORMOVE 0x00
#define LCD_MOVERIGHT 0x04
#define LCD_MOVELEFT 0x00

// flags for function set
#define LCD_8BITMODE 0x10
#define LCD_4BITMODE 0x00
#define LCD_2LINE 0x08
#define LCD_1LINE 0x00
#define LCD_5x10DOTS 0x04
#define LCD_5x8DOTS 0x00
#define LCD_PAGE2 0x02
#define LCD_PAGE1 0x00


/*
 * PB0 - data
 * PB1 - latch
 * PB2 - clock
 */

#define SETUP_CONTROL_SHREG_PINS \
					PORTB &= 0xF8; \
					DDRB |= 0x07;

#define LATCH_SHREG \
					PORTB |= 0x02; \
					delayMicroseconds(10); \
					PORTB &= 0xFD;

#define SEND_HI \
					PORTB |= 0x01; \
					PORTB |= 0x04; \
					delayMicroseconds(10); \
					PORTB &= 0xFA;

#define SEND_LOW \
					PORTB &= 0xFE; \
					PORTB |= 0x04; \
					delayMicroseconds(10); \
					PORTB &= 0xFB;


#ifdef __cplusplus
extern "C" {
#endif


void LCD_begin();
void LCD_printChar(char ch);
void LCD_printNumber(uint8_t n);
void LCD_printTime(uint8_t m, uint8_t s);
void LCD_printUTF8String(const char *utf8String);
void LCD_printUTF8Char(uint16_t ch);
void LCD_noBlink();
void LCD_blink();
void LCD_clear();
void LCD_setCursor(uint8_t col, uint8_t row);


#ifdef __cplusplus
}
#endif

#endif /* LCD3WI_H_ */
