/*
 * main.cpp
 *
 *  Created on: 8 сент. 2016 г.
 *      Author: alexandr
 */


#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

#include "LCD3WI.h"
#include "ADCKeyboard.h"
#include "Message.h"

#include "Utils.h"
#include "Config.h"


struct Config config EEMEM =
{
	0,	// Language
	0,	// Cancel Button
	{	// Delay
			0,	// Minute
			0	// Second
	},
	{
		{	// Preset #0
				{ 0 }, 	// name
				{	// Time
						0,	// Minute
						0	// Second
				}
		},
		{	// Preset #1
				{ 0 }, 	// name
				{	// Time
						0,	// Minute
						0	// Second
				}
		},
		{	// Preset #2
				{ 0 }, 	// name
				{	// Time
						0,	// Minute
						0	// Second
				}
		},
		{	// Preset #3
				{ 0 }, 	// name
				{	// Time
						0,	// Minute
						0	// Second
				}
		},
		{	// Preset #4
				{ 0 }, 	// name
				{	// Time
						0,	// Minute
						0	// Second
				}
		}
	}
};


Message message;


void doLEDs(struct LEDTime time);

void doInitialState(ADC_KEY keyData);
	void doPresetState(ADC_KEY keyData);
		void doPresetNameState(ADC_KEY keyData);
		void doPresetTimeState(ADC_KEY keyData);
	void doLEDState(ADC_KEY keyData);
	void doSettingsState(ADC_KEY keyData);
		void doLanguageState(ADC_KEY keyData);
		void doDelayState(ADC_KEY keyData);

void printWelcomeState();
void printPresetName(const uint16_t *name, bool prefix);

void printToLCD(MessageCodes code);

void (*currentState)(ADC_KEY) = NULL;


struct SubState {
	uint8_t elements;
	uint8_t topElementIndex;
	uint8_t selectedRow;
};

#define SUBSTATES_COUNT			4
#define SUBSTATE_INITIAL_INDEX	0
#define SUBSTATE_PRESET_INDEX	1
#define SUBSTATE_SETTINGS_INDEX	2
#define SUBSTATE_LANGUAGE_INDEX	3
#define INITIAL_ELEMENTS_COUNT	(PRESETS_COUNT + 2)
#define PRESET_ELEMENTS_COUNT	4
#define SETTINGS_ELEMENTS_COUNT	2
#define LANGUAGE_ELEMENTS_COUNT	2

struct SubState states[SUBSTATES_COUNT] =
			{
					{ INITIAL_ELEMENTS_COUNT, 0, 0 },
					{ PRESET_ELEMENTS_COUNT, 0, 0 },
					{ SETTINGS_ELEMENTS_COUNT, 0, 0 },
					{ LANGUAGE_ELEMENTS_COUNT, 0, 0 }
			};


volatile uint8_t selectedCharPos = 0;
volatile uint8_t selectedTimePos = 0;
volatile uint8_t selectedTimeDigit = 0;
volatile int8_t selectedPreset = -1;
volatile uint8_t language = 0;
struct Preset currentPreset;

int main()
{
	PORTB &= 0xF0;
	DDRB |= 0x0F;

	ADCSRA = 0;

	sei();

	LCD_begin();

	language = eeprom_read_byte(&config.language);

	printWelcomeState();

	adcKeyboard.enable();

	while(true) {
		if (adcKeyboard.isKeyDataAvailable()) {
			ADC_KEY keyData = adcKeyboard.getKeyData();
			if (!currentState) {
				currentState = doInitialState;
				keyData = ADC_KEY::KEY_INIT;
			}
			currentState(keyData);
		}
	}

}

bool doTime(MessageCodes code, struct LEDTime time)
{
	LCD_clear();
	LCD_setCursor(0, 0);
	printToLCD(code);
	for (; time.minute >= 0; time.minute--) {
		for (; time.second >= 0; time.second--) {
			LCD_setCursor(0, 1);
			LCD_printTime(time.minute, time.second);
			if (adcKeyboard.isKeyDataAvailable()) {
				adcKeyboard.getKeyData();
				return false;
			}
			if ((time.minute > 0) || (time.second > 0)) {
				delayMilliseconds(999);
			}
		}
		time.second = 59;
	}
	return true;
}

void doLEDs(struct LEDTime time)
{
	LEDTime delay;
	eeprom_read_block(&delay, &config.delay, sizeof(struct LEDTime));
	if (doTime(MSG_DELAY, delay)) {
		PORTB |= 0x08;	// Torn LEDs on
		doTime(MSG_LEDs, time);
		PORTB &= 0xF7;	// Torn LEDs off
	}
	currentState(KEY_INIT);
}

void printElementsState(uint8_t statesIndex, void (*printStateElement) (uint8_t))
{
	LCD_clear();
	for (uint8_t i = 0; i < 2; i++) {
		LCD_setCursor(0, i);
		LCD_printChar(i == states[statesIndex].selectedRow ? '>' : ' ');
		printStateElement((states[statesIndex].topElementIndex + i) % states[statesIndex].elements);
		LCD_printChar(i == states[statesIndex].selectedRow ? '<' : ' ');
	}
}

void changeState(void (*nextState)(ADC_KEY))
{
	currentState = nextState;
	if (currentState == NULL) {
		printWelcomeState();
	} else {
		currentState(KEY_INIT);
	}
}

void doElementsState(ADC_KEY keyData,
						uint8_t statesIndex,
						void (*printStateElement) (uint8_t),
						void (*doOk)(uint8_t),
						void (*prevState)(ADC_KEY)
						)
{
	switch(keyData) {
	case KEY_INIT:
		break;
	case KEY_UP:
		if (states[statesIndex].selectedRow == 0) {
			states[statesIndex].topElementIndex = (states[statesIndex].topElementIndex + states[statesIndex].elements - 1)
														% states[statesIndex].elements;
		} else {
			states[statesIndex].selectedRow--;
		}
		break;
	case KEY_DOWN:
		if (states[statesIndex].selectedRow == 1) {
			states[statesIndex].topElementIndex = (states[statesIndex].topElementIndex + 1)
														% states[statesIndex].elements;
		} else {
			states[statesIndex].selectedRow++;
		}
		break;
	case KEY_OK:
		doOk((states[statesIndex].topElementIndex + states[statesIndex].selectedRow) % states[statesIndex].elements);
		return;
	case KEY_CANCEL:
		changeState(prevState);
		return;
	}
	printElementsState(statesIndex, printStateElement);
}

void doTimeState(ADC_KEY keyData, MessageCodes code, uint8_t stateIndex)
{
	switch(keyData) {
	case KEY_INIT:
		selectedCharPos = message.getMessageSize(code) + 4;
		selectedTimeDigit = currentPreset.time.minute;
		selectedTimePos = 0;
		LCD_blink();
		break;
	case KEY_DOWN:
		selectedTimeDigit = (selectedTimeDigit + 60 - 1) % 60;
		LCD_setCursor(selectedCharPos - 1, states[stateIndex].selectedRow);
		LCD_printNumber(selectedTimeDigit);
		break;
	case KEY_UP:
		selectedTimeDigit = (selectedTimeDigit + 1) % 60;
		LCD_setCursor(selectedCharPos - 1, states[stateIndex].selectedRow);
		LCD_printNumber(selectedTimeDigit);
		break;
	case KEY_OK:
		if (selectedTimePos) {
			LCD_noBlink();
			currentPreset.time.second = selectedTimeDigit;
			selectedTimePos++;
			return;
		} else {
			currentPreset.time.minute = selectedTimeDigit;
			selectedTimeDigit = currentPreset.time.second;
			selectedCharPos += 3;
			selectedTimePos++;
		}
		break;
	case KEY_CANCEL:
		LCD_noBlink();
		return;
	}
	LCD_setCursor(selectedCharPos, states[stateIndex].selectedRow);
}

// ======================== WELCOME STATE

void printWelcomeState()
{
	LCD_clear();
	LCD_setCursor(0, 0);
	printToLCD(MessageCodes::MSG_LEDs);
	LCD_printChar(' ');
	printToLCD(MessageCodes::MSG_ENGINE);
	LCD_setCursor(0, 1);
	LCD_printUTF8String("0.1");
}

// ==================================   INITIAL STATE

void doInitialStateOk(uint8_t selected)
{
	if (selected < PRESETS_COUNT) {
		selectedPreset = selected;
		eeprom_read_block(&currentPreset, &config.presets[selectedPreset], sizeof(struct Preset));
		currentState = doPresetState;
	} else switch (selected) {
	case PRESETS_COUNT:
		currentState = doLEDState;
		break;
	case PRESETS_COUNT + 1:
		currentState = doSettingsState;
		break;
	}
	currentState(KEY_INIT);
}

void printInitialStateElement(uint8_t element)
{
	if (element < PRESETS_COUNT) {
		eeprom_read_block(&currentPreset, &config.presets[element], sizeof(struct Preset));
		selectedPreset = element;
		printPresetName(currentPreset.name, true);
	} else switch (element) {
	case PRESETS_COUNT:
		printToLCD(MessageCodes::MSG_LEDs);
		LCD_printChar(' ');
		printToLCD(MessageCodes::MSG_ON);
		break;
	case PRESETS_COUNT + 1:
		printToLCD(MessageCodes::MSG_SETTINGS);
		break;
	}
}

void doInitialState(ADC_KEY keyData)
{
	doElementsState(keyData, SUBSTATE_INITIAL_INDEX, printInitialStateElement, doInitialStateOk, NULL);
}

// ==================================   PRESET STATE

void printPresetName(const uint16_t *name, bool prefix)
{
	uint16_t ch = name[0];
	if (ch) {
		if (prefix) {
			LCD_printNumber(selectedPreset);
			LCD_printChar('.');
		}
		for (int i = 1; (i < 16) && ch; i++) {
			LCD_printUTF8Char(ch);
			ch = name[i];
		}
	} else {
		printToLCD(MessageCodes::MSG_PRESET);
		LCD_printUTF8String(" #");
		LCD_printNumber(selectedPreset);
	}
}

void printElementTime(MessageCodes code)
{
	printToLCD(code);
	LCD_printUTF8String(": ");
	LCD_printTime(currentPreset.time.minute, currentPreset.time.second);
}

void printPresetStateElement(uint8_t presetSubState)
{
	switch (presetSubState) {
	case 0:
		printPresetName(currentPreset.name, false);
		break;
	case 1:
		printElementTime(MessageCodes::MSG_TIME);
		break;
	case 2:
		printToLCD(MessageCodes::MSG_SAVE);
		break;
	case 3:
		printToLCD(MessageCodes::MSG_START);
		break;
	}
}

void doPresetStateOk(uint8_t selected)
{
	switch (selected) {
	case 0:
		currentState = doPresetNameState;
		currentState(KEY_INIT);
		break;
	case 1:
		currentState = doPresetTimeState;
		currentState(KEY_INIT);
		break;
	case 2:
		eeprom_update_block(&currentPreset, &config.presets[selectedPreset], sizeof(struct Preset));
		break;
	case 3:
		doLEDs(currentPreset.time);
		break;
	}
}

void doPresetState(ADC_KEY keyData)
{
	doElementsState(keyData, SUBSTATE_PRESET_INDEX, printPresetStateElement, doPresetStateOk, doInitialState);
}


#define CHARSET_SIZE	69


uint16_t fixNextChar(uint16_t ch)
{
	switch (ch) {
	case 0x0001:
	case 0xFFFF:
		if (language)
			return 0xD090;
		else
			return 0x0041;
	case 0xD096:	// 'Ж'
	case 0xD095:	// 'Е'
		return 0xD081;	// 'Ё'
	case 0xD080:	// < 'Ё'
		return 0xD095;	// 'Е'
	case 0xD082:	// > 'Ё'
		return 0xD096;	// 'Ж'
	case 0xD0B0:	// > 'Я'
		return 0x0020;	// ' '
	case 0x001F:	// < ' '
		return 0xD0AF;	// 'Я'
	case 0x005B:	// > 'Z'
		return 0xAC;	// 'Enter'
	case 0xD08F:	// < 'А'
		return 0xAC;	// 'Enter'
	case 0x00AD:	// > 'Enter'
		return 0xD090;	// 'А'
	case 0x00AB:	// < 'Enter'
		return 0x005A;	// 'Z'
	}
	return ch;
}

void doPresetNameState(ADC_KEY keyData)
{
	switch(keyData) {
	case KEY_INIT:
		selectedCharPos = 0;
		LCD_setCursor(1 + selectedCharPos, states[SUBSTATE_PRESET_INDEX].selectedRow);
		if (!currentPreset.name[selectedCharPos]) {
			if (language)
				currentPreset.name[selectedCharPos] = 0xD090;
			else
				currentPreset.name[selectedCharPos] = 0x0041;
			LCD_printChar('A');
			for (int i = 2; i < 16; i++) {
				LCD_printChar(' ');
			}
		}
		LCD_blink();
		break;
	case KEY_DOWN:
		currentPreset.name[selectedCharPos] = fixNextChar(currentPreset.name[selectedCharPos] - 1);
		LCD_printUTF8Char(currentPreset.name[selectedCharPos]);
		break;
	case KEY_UP:
		currentPreset.name[selectedCharPos] = fixNextChar(currentPreset.name[selectedCharPos] + 1);
		LCD_printUTF8Char(currentPreset.name[selectedCharPos]);
		break;
	case KEY_OK:
		if (currentPreset.name[selectedCharPos] == 0xAC) {
			currentPreset.name[selectedCharPos] = 0;
		}
		if (currentPreset.name[selectedCharPos]) {
			selectedCharPos++;
		} else {
			LCD_noBlink();
			changeState(doPresetState);
			return;
		}
		break;
	case KEY_CANCEL:
		LCD_noBlink();
		eeprom_read_block(currentPreset.name, &config.presets[selectedPreset].name, 32);
		changeState(doPresetState);
		return;
	}
	LCD_setCursor(1 + selectedCharPos, states[SUBSTATE_PRESET_INDEX].selectedRow);
}

void doPresetTimeState(ADC_KEY keyData)
{
	doTimeState(keyData, MessageCodes::MSG_TIME, SUBSTATE_PRESET_INDEX);
	if (keyData == KEY_CANCEL) {
		eeprom_read_block(&currentPreset.time, &config.presets[selectedPreset].time, sizeof(struct LEDTime));
		changeState(doPresetState);
	} else if ((keyData == KEY_OK) && (selectedTimePos > 1)) {
		changeState(doPresetState);
	}
}

// ==================================  LED STATE

void doLEDState(ADC_KEY keyData)
{
	if (keyData == KEY_INIT) {
		LCD_clear();
		LCD_setCursor(0, 0);
		printToLCD(MessageCodes::MSG_LEDs);
		LCD_printChar(' ');
		printToLCD(MessageCodes::MSG_OFF);
		PORTB |= 0x08;
	} else {
		PORTB &= 0xF7;	// Torn LEDs off
		changeState(doInitialState);
	}
}

// =================================  SETTINGS STATE

void printSettingsStateElement(uint8_t presetSubState)
{
	switch (presetSubState) {
	case 0:
		printToLCD(MessageCodes::MSG_LANGUAGE);
		break;
	case 1:
		printElementTime(MessageCodes::MSG_DELAY);
		break;
	}
}

void doSettingsStateOk(uint8_t selected)
{
	switch (selected) {
	case 0:
		currentState = doLanguageState;
		break;
	case 1:
		currentState = doDelayState;
		break;
	}
	currentState(KEY_INIT);
}

void doSettingsState(ADC_KEY keyData)
{
	eeprom_read_block(&currentPreset.time, &config.delay, sizeof(struct LEDTime));
	doElementsState(keyData, SUBSTATE_SETTINGS_INDEX, printSettingsStateElement, doSettingsStateOk, doInitialState);
}

// ================================   LANGUAGE

void printLanguageStateElement(uint8_t languaugeSubState)
{
	switch (languaugeSubState) {
	case 0:
		if (language) {
			LCD_printChar(' ');
		} else {
			LCD_printChar('*');
		}
		printToLCD(MessageCodes::MSG_ENGLISH);
		break;
	case 1:
		if (language) {
			LCD_printChar('*');
		} else {
			LCD_printChar(' ');
		}
		printToLCD(MessageCodes::MSG_RUSSIAN);
		break;
	}
}

void doChangeLanguage(uint8_t lang)
{
	language = lang;
	LCD_setCursor(1, states[SUBSTATE_LANGUAGE_INDEX].selectedRow);
	LCD_printChar('*');
	if (states[SUBSTATE_LANGUAGE_INDEX].topElementIndex == 0) {
		LCD_setCursor(1, (states[SUBSTATE_LANGUAGE_INDEX].selectedRow + 1) % 2);
		LCD_printChar(' ');
	}
}

void doLanguageStateOk(uint8_t selected)
{
	doChangeLanguage(selected);
	eeprom_update_byte(&config.language, language);
}

void doLanguageState(ADC_KEY keyData)
{
	doElementsState(keyData, SUBSTATE_LANGUAGE_INDEX, printLanguageStateElement, doLanguageStateOk, doSettingsState);
}

// ================================   DELAY

void doDelayState(ADC_KEY keyData)
{
	doTimeState(keyData, MessageCodes::MSG_DELAY, SUBSTATE_SETTINGS_INDEX);
	if (keyData == KEY_CANCEL) {
		eeprom_read_block(&currentPreset.time, &config.delay, sizeof(struct LEDTime));
		changeState(doSettingsState);
	} else if ((keyData == KEY_OK) && (selectedTimePos > 1)) {
		eeprom_update_block(&currentPreset.time, &config.delay, sizeof(struct LEDTime));
		changeState(doSettingsState);
	}
}


char messageBuffer[34];

void printToLCD(MessageCodes code)
{
	message.getMessage(code, messageBuffer, 32);
	LCD_printUTF8String(messageBuffer);
}

